@@ -42,7 +42,7 @@ def create_enum(data):
                 key = item["@id"].split(":")[-1]
             if key in items:
                 key = item["@id"].replace(":", "_")
-            items[key.replace("-", "_")] = item["@id"]
+            items[key.replace("-", "_").replace(".", "")] = item["@id"]
     if klass is None or len(items) == 0:
         raise ValueError(f"Could not generate a klass or items from {data}")
     newklass = Enum(klass, items)
@@ -113,7 +113,24 @@ def diff_models(model1, model2):
             print(f"{field} is different")
 
 
+def _sanitize(o):
+    if isinstance(o, dict):
+        return {_sanitize(k): _sanitize(v) for k, v in o.items()}
+    elif isinstance(o, (set, tuple, list)):
+        return type(o)(_sanitize(x) for x in o)
+    elif isinstance(o, Enum):
+        return o.value
+    return o
+
+
+class HandleKeyEnumEncoder(json.JSONEncoder):
+    def encode(self, o):
+        return super().encode(_sanitize(o))
+
+
 class DandiBaseModel(BaseModel):
+    id: Optional[str] = Field(description="Uniform resource identifier", readOnly=True)
+
     @classmethod
     def unvalidated(__pydantic_cls__: Type[BaseModel], **data: Any) -> BaseModel:
         """Allow model to be returned without validation"""
@@ -151,6 +168,8 @@ class DandiBaseModel(BaseModel):
             for prop, value in schema.get("properties", {}).items():
                 if value.get("title") is None or value["title"] == prop.title():
                     value["title"] = name2title(prop)
+                if value.get("format", None) == "uri":
+                    value["maxLength"] = 1000
                 allOf = value.get("allOf")
                 anyOf = value.get("anyOf")
                 items = value.get("items")
@@ -259,7 +278,7 @@ class SpeciesType(TypeModel):
 class Disorder(TypeModel):
     """Biolink, SNOMED, or other identifier for disorder studied"""
 
-    dxdate: Optional[List[date]] = Field(
+    dxdate: Optional[List[Union[date, datetime]]] = Field(
         None,
         title="Dates of diagnosis",
         description="Dates of diagnosis",
@@ -269,10 +288,10 @@ class Disorder(TypeModel):
     schemaKey: Literal["Disorder"] = Field("Disorder", readOnly=True)
 
 
-class ModalityType(TypeModel):
-    """Identifier for modality used"""
+class ApproachType(TypeModel):
+    """Identifier for approach used"""
 
-    schemaKey: Literal["ModalityType"] = Field("ModalityType", readOnly=True)
+    schemaKey: Literal["ApproachType"] = Field("ApproachType", readOnly=True)
 
 
 class MeasurementTechniqueType(TypeModel):
@@ -360,7 +379,9 @@ class Person(Contributor):
         nskey="schema",
     )
     name: str = Field(
-        description="Use the format: lastname, firstname ...", nskey="schema"
+        description="Use the format: familyname, given names ...",
+        nskey="schema",
+        examples=["Lovelace, Augusta Ada", "Smith, John", "Chan, Kong-sang"],
     )
     affiliation: List[Organization] = Field(
         None,
@@ -391,6 +412,23 @@ class Software(DandiBaseModel):
     }
 
 
+class Agent(DandiBaseModel):
+    identifier: Optional[Identifier] = Field(
+        None,
+        title="Identifier",
+        description="Identifier for an agent",
+        nskey="schema",
+    )
+    name: str = Field(nskey="schema")
+    url: Optional[HttpUrl] = Field(None, nskey="schema")
+    schemaKey: Literal["Software"] = Field("Agent", readOnly=True)
+
+    _ldmeta = {
+        "rdfs:subClassOf": ["prov:Agent"],
+        "nskey": "dandi",
+    }
+
+
 class EthicsApproval(DandiBaseModel):
     """Information about ethics committee approval for project"""
 
@@ -467,7 +505,7 @@ class AssetsSummary(DandiBaseModel):
     # stats which are not stats
     numberOfBytes: int = Field(readOnly=True, sameas="schema:contentSize")
     numberOfFiles: int = Field(readOnly=True)  # universe
-    numberOfSubjects: int = Field(readOnly=True)  # NWB + BIDS
+    numberOfSubjects: Optional[int] = Field(None, readOnly=True)  # NWB + BIDS
     numberOfSamples: Optional[int] = Field(None, readOnly=True)  # more of NWB
     numberOfCells: Optional[int] = Field(None, readOnly=True)
 
@@ -475,7 +513,7 @@ class AssetsSummary(DandiBaseModel):
         readOnly=True
     )  # TODO: types of things NWB, BIDS
     # Web UI: icons per each modality?
-    modality: List[ModalityType] = Field(
+    approach: List[ApproachType] = Field(
         readOnly=True
     )  # TODO: types of things, BIDS etc...
     # Web UI: could be an icon with number, which if hovered on  show a list?
@@ -490,155 +528,33 @@ class AssetsSummary(DandiBaseModel):
     }
 
 
-class Digest(DandiBaseModel):
-    """Information about the crytographic checksum of the item."""
-
-    value: str = Field(nskey="schema")
-    cryptoType: DigestType = Field(
-        title="Cryptographic method used",
-        description="Which cryptographic checksum is used",
-        nskey="dandi",
-    )
-
-    _ldmeta = {
-        "rdfs:subClassOf": ["schema:Thing", "prov:Entity"],
-        "rdfs:label": "Cryptographic checksum information",
-        "nskey": "dandi",
-    }
-
-
-class BioSample(DandiBaseModel):
-    """Description about the sample that was studied"""
-
-    identifier: Identifier = Field(nskey="schema")
-    sampleType: Optional[SampleType] = Field(
-        None, description="OBI based identifier for the sample used", nskey="dandi"
-    )
-    assayType: Optional[List[AssayType]] = Field(
-        None, description="OBI based identifier for the assay(s) used", nskey="dandi"
-    )
-    anatomy: Optional[List[Anatomy]] = Field(
-        None,
-        description="UBERON based identifier for what organ the sample belongs "
-        "to. Use the most specific descriptor.",
-        nskey="dandi",
-    )
-
-    wasDerivedFrom: Optional[List["BioSample"]] = Field(None, nskey="prov")
-    sameAs: Optional[List[Identifier]] = Field(None, nskey="schema")
-    hasMember: Optional[List[Identifier]] = Field(None, nskey="prov")
-
-    schemaKey: Literal["BioSample"] = Field("BioSample", readOnly=True)
-
-    _ldmeta = {
-        "rdfs:subClassOf": ["schema:Thing", "prov:Entity"],
-        "rdfs:label": "Information about the biosample.",
-        "nskey": "dandi",
-    }
-
-
-BioSample.update_forward_refs()
-
-
-class RelatedParticipant(DandiBaseModel):
-    identifier: Optional[Identifier] = Field(None, nskey="schema")
-    name: Optional[str] = Field(None, title="A title of the resource", nskey="schema")
-    url: Optional[HttpUrl] = Field(None, title="URL of the resource", nskey="schema")
-    relation: ParticipantRelationType = Field(
-        title="Choose a relation satisfying: Participant <relation> relatedParticipant",
-        description="Indicates how the current participant is related to the other participant "
-        "This relation should satisfy: Participant <relation> relatedParticipant",
-        nskey="dandi",
-    )
-
-    _ldmeta = {
-        "rdfs:subClassOf": ["schema:CreativeWork", "prov:Entity"],
-        "rdfs:comment": "Another participant related to the participant (e.g., another "
-        "parent, sibling, child)",
-        "nskey": "dandi",
-    }
-
-
-class Participant(DandiBaseModel):
-    """Description about the sample that was studied"""
-
-    identifier: Identifier = Field(nskey="schema")
-    source_id: Optional[Identifier] = Field(None, nskey="dandi")
-
-    strain: Optional[StrainType] = Field(
-        None, description="Identifier for the strain of the sample", nskey="dandi"
-    )
-    cellLine: Optional[Identifier] = Field(
-        None, description="Cell line associated with the sample", nskey="dandi"
-    )
-    vendor: Optional[Organization] = Field(None, nskey="dandi")
-    age: Optional[PropertyValue] = Field(
-        None,
-        description="A representation of age using ISO 8601 duration. This "
-        "should include a valueReference if anything other than "
-        "date of birth is used.",
-        nskey="dandi",
-        rangeIncludes="schema:Duration",
-    )
-    sex: Optional[SexType] = Field(
-        None,
-        description="OBI based identifier for sex of the sample if available",
-        nskey="dandi",
-    )
-    genotype: Optional[Identifier] = Field(
-        None, description="Genotype descriptor of biosample if available", nskey="dandi"
-    )
-    species: Optional[SpeciesType] = Field(
-        None,
-        description="An identifier indicating the taxonomic classification of the biosample",
-        nskey="dandi",
-    )
-    disorder: Optional[List[Disorder]] = Field(
-        None,
-        description="Any current diagnosed disease or disorder associated with the sample",
-        nskey="dandi",
-    )
-    relatedParticipant: Optional[List[RelatedParticipant]] = Field(None, nskey="dandi")
-
-    schemaKey: Literal["Participant"] = Field("Participant", readOnly=True)
-
-    _ldmeta = {
-        "rdfs:subClassOf": ["schema:Thing", "prov:Entity"],
-        "rdfs:label": "Information about the participant.",
-        "nskey": "dandi",
-    }
-
-
 class Activity(DandiBaseModel):
     """Information about the Project activity"""
 
     identifier: Optional[Identifier] = Field(None, nskey="schema")
     name: str = Field(
         title="Title",
-        description="The name of the item.",
+        description="The name of the activity.",
         max_length=150,
         nskey="schema",
     )
     description: Optional[str] = Field(
-        None, description="A description of the item.", nskey="schema"
+        None, description="The description of the activity.", nskey="schema"
     )
-    startDate: Optional[date] = Field(None, nskey="schema")
-    endDate: Optional[date] = Field(None, nskey="schema")
+    startDate: Optional[datetime] = Field(None, nskey="schema")
+    endDate: Optional[datetime] = Field(None, nskey="schema")
 
     # isPartOf: Optional["Activity"] = Field(None, nskey="schema")
     # hasPart: Optional["Activity"] = Field(None, nskey="schema")
-    wasAssociatedWith: Optional[List[Union[Person, Organization, Software]]] = Field(
-        None, nskey="prov"
-    )
+    wasAssociatedWith: Optional[
+        List[Union[Person, Organization, Software, Agent]]
+    ] = Field(None, nskey="prov")
 
     schemaKey: Literal["Activity"] = Field("Activity", readOnly=True)
 
     _ldmeta = {"rdfs:subClassOf": ["prov:Activity", "schema:Thing"], "nskey": "dandi"}
 
 
-# Activity.update_forward_refs()
-
-
 class Project(Activity):
     name: str = Field(
         title="Title",
@@ -665,8 +581,152 @@ class Session(Activity):
     schemaKey: Literal["Session"] = Field("Session", readOnly=True)
 
 
+class PublishActivity(Activity):
+    schemaKey: Literal["PublishActivity"] = Field("PublishActivity", readOnly=True)
+
+
+class Locus(DandiBaseModel):
+    identifier: Union[Identifier, List[Identifier]] = Field(
+        description="Identifier for genotyping locus"
+    )
+    locus_type: str = Field()
+    symbol: str = Field()
+    schemaKey: Literal["Locus"] = Field("Locus", readOnly=True)
+    _ldmeta = {"nskey": "dandi"}
+
+
+class Allele(DandiBaseModel):
+    identifier: Union[Identifier, List[Identifier]] = Field(
+        description="Identifier for genotyping allele"
+    )
+    allele_type: str = Field()
+    symbol: str = Field()
+    schemaKey: Literal["Allele"] = Field("Allele", readOnly=True)
+    _ldmeta = {"nskey": "dandi"}
+
+
+class GenotypeInfo(DandiBaseModel):
+    locus: Locus = Field(description="Locus at which information was extracted")
+    alleles: List[Allele] = Field(
+        max_items=3, description="Information about one allele"
+    )
+    wasGeneratedBy: Optional[List["Session"]] = Field(None, nskey="prov")
+    schemaKey: Literal["GenotypeInfo"] = Field("GenotypeInfo", readOnly=True)
+    _ldmeta = {"nskey": "dandi"}
+
+
+class RelatedParticipant(DandiBaseModel):
+    identifier: Optional[Identifier] = Field(None, nskey="schema")
+    name: Optional[str] = Field(None, title="A name of the Participant", nskey="schema")
+    url: Optional[HttpUrl] = Field(
+        None, title="URL of the related participant", nskey="schema"
+    )
+    relation: ParticipantRelationType = Field(
+        title="Choose a relation satisfying: Participant <relation> relatedParticipant",
+        description="Indicates how the current participant is related to the other participant "
+        "This relation should satisfy: Participant <relation> relatedParticipant",
+        nskey="dandi",
+    )
+
+    _ldmeta = {
+        "rdfs:subClassOf": ["schema:CreativeWork", "prov:Entity"],
+        "rdfs:comment": "Another participant related to the participant (e.g., another "
+        "parent, sibling, child)",
+        "nskey": "dandi",
+    }
+
+
+class Participant(DandiBaseModel):
+    """Description about the sample that was studied"""
+
+    identifier: Optional[Identifier] = Field(nskey="schema")
+    altName: Optional[List[Identifier]] = Field(None, nskey="dandi")
+
+    strain: Optional[StrainType] = Field(
+        None, description="Identifier for the strain of the sample", nskey="dandi"
+    )
+    cellLine: Optional[Identifier] = Field(
+        None, description="Cell line associated with the sample", nskey="dandi"
+    )
+    vendor: Optional[Organization] = Field(None, nskey="dandi")
+    age: Optional[PropertyValue] = Field(
+        None,
+        description="A representation of age using ISO 8601 duration. This "
+        "should include a valueReference if anything other than "
+        "date of birth is used.",
+        nskey="dandi",
+        rangeIncludes="schema:Duration",
+    )
+
+    sex: Optional[SexType] = Field(
+        None,
+        description="OBI based identifier for sex of the sample if available",
+        nskey="dandi",
+    )
+    genotype: Optional[Union[List[GenotypeInfo], Identifier]] = Field(
+        None, description="Genotype descriptor of biosample if available", nskey="dandi"
+    )
+    species: Optional[SpeciesType] = Field(
+        None,
+        description="An identifier indicating the taxonomic classification of the biosample",
+        nskey="dandi",
+    )
+    disorder: Optional[List[Disorder]] = Field(
+        None,
+        description="Any current diagnosed disease or disorder associated with the sample",
+        nskey="dandi",
+    )
+
+    relatedParticipant: Optional[List[RelatedParticipant]] = Field(None, nskey="dandi")
+    sameAs: Optional[List[Identifier]] = Field(None, nskey="schema")
+
+    schemaKey: Literal["Participant"] = Field("Participant", readOnly=True)
+
+    _ldmeta = {
+        "rdfs:subClassOf": ["prov:Agent"],
+        "rdfs:label": "Information about the participant.",
+        "nskey": "dandi",
+    }
+
+
+class BioSample(DandiBaseModel):
+    """Description of the sample that was studied"""
+
+    identifier: Optional[Identifier] = Field(nskey="schema")
+    sampleType: Optional[SampleType] = Field(
+        None, description="OBI based identifier for the sample used", nskey="dandi"
+    )
+    assayType: Optional[List[AssayType]] = Field(
+        None, description="OBI based identifier for the assay(s) used", nskey="dandi"
+    )
+    anatomy: Optional[List[Anatomy]] = Field(
+        None,
+        description="UBERON based identifier for what organ the sample belongs "
+        "to. Use the most specific descriptor.",
+        nskey="dandi",
+    )
+
+    wasDerivedFrom: Optional[List["BioSample"]] = Field(None, nskey="prov")
+    wasAttributedTo: Optional[List[Participant]] = Field(
+        None, description="Participant(s) to which this sample belongs to", nskey="prov"
+    )
+    sameAs: Optional[List[Identifier]] = Field(None, nskey="schema")
+    hasMember: Optional[List[Identifier]] = Field(None, nskey="prov")
+
+    schemaKey: Literal["BioSample"] = Field("BioSample", readOnly=True)
+
+    _ldmeta = {
+        "rdfs:subClassOf": ["schema:Thing", "prov:Entity"],
+        "rdfs:label": "Information about the biosample.",
+        "nskey": "dandi",
+    }
+
+
+BioSample.update_forward_refs()
+
+
 class Identifiable(DandiBaseModel):
-    identifier: Identifier = Field(readOnly=True, nskey="schema")
+    identifier: Optional[Identifier] = Field(readOnly=True, nskey="schema")
 
 
 class CommonModel(DandiBaseModel):
@@ -735,10 +795,10 @@ class CommonModel(DandiBaseModel):
     def json_dict(self):
         """
         Recursively convert the instance to a `dict` of JSONable values,
-        including converting enum values to strings.  Unset and `None` fields
+        including converting enum values to strings.  `None` fields
         are omitted.
         """
-        return json.loads(self.json(exclude_unset=True, exclude_none=True))
+        return json.loads(self.json(exclude_none=True, cls=HandleKeyEnumEncoder))
 
 
 class DandisetMeta(CommonModel, Identifiable):
@@ -754,6 +814,8 @@ class DandisetMeta(CommonModel, Identifiable):
             raise ValueError("At least one contributor must have role ContactPerson")
         return values
 
+    id: str = Field(description="Uniform resource identifier", readOnly=True)
+
     identifier: DANDI = Field(
         readOnly=True,
         title="Dandiset identifier",
@@ -767,7 +829,6 @@ class DandisetMeta(CommonModel, Identifiable):
         max_length=150,
         nskey="schema",
     )
-
     description: str = Field(
         description="A description of the Dandiset", max_length=3000, nskey="schema"
     )
@@ -777,6 +838,12 @@ class DandisetMeta(CommonModel, Identifiable):
         nskey="schema",
         min_items=1,
     )
+    dateCreated: Optional[datetime] = Field(
+        nskey="schema", title="Dandiset creation date and time", readOnly=True
+    )
+    dateModified: Optional[datetime] = Field(
+        nskey="schema", title="Last modification date and time", readOnly=True
+    )
 
     citation: TempOptional[str] = Field(readOnly=True, nskey="schema")
 
@@ -810,15 +877,6 @@ class DandisetMeta(CommonModel, Identifiable):
     }
 
 
-class PublishedDandisetMeta(DandisetMeta):
-    publishedBy: HttpUrl = Field(
-        description="The URL should contain the provenance of the publishing process.",
-        readOnly=True,
-        nskey="dandi",
-    )  # TODO: formalize "publish" activity to at least the Actor
-    datePublished: date = Field(readOnly=True, nskey="schema")
-
-
 class BareAssetMeta(CommonModel):
     """Metadata used to describe an asset anywhere (local or server).
 
@@ -835,7 +893,7 @@ class BareAssetMeta(CommonModel):
     encodingFormat: Union[HttpUrl, str] = Field(
         title="File Encoding Format", nskey="schema"
     )
-    digest: List[Digest] = Field(nskey="dandi", default_factory=list)
+    digest: Dict[DigestType, str] = Field(default_factory=dict)
     dateModified: Optional[datetime] = Field(
         nskey="schema", title="Asset (file or metadata) modification date and time"
     )
@@ -853,7 +911,7 @@ class BareAssetMeta(CommonModel):
     sameAs: Optional[List[HttpUrl]] = Field(None, nskey="schema")
 
     # TODO
-    modality: Optional[List[ModalityType]] = Field(None, readOnly=True, nskey="dandi")
+    approach: Optional[List[ApproachType]] = Field(None, readOnly=True, nskey="dandi")
     measurementTechnique: Optional[List[MeasurementTechniqueType]] = Field(
         None, readOnly=True, nskey="schema"
     )
@@ -882,19 +940,38 @@ class BareAssetMeta(CommonModel):
 class AssetMeta(BareAssetMeta, Identifiable):
     """Metadata used to describe an asset on the server."""
 
-    identifier: UUID4 = Field(readOnly=True, nskey="schema")
-
+    id: Optional[str] = Field(readOnly=True, description="Uniform resource identifier")
+    identifier: Optional[UUID4] = Field(readOnly=True, nskey="schema")
     # on publish or set by server
     contentUrl: Optional[List[HttpUrl]] = Field(None, readOnly=True, nskey="schema")
 
 
-class PublishedAssetMeta(AssetMeta):
-    publishedBy: HttpUrl = Field(
+class Publishable(DandiBaseModel):
+    id: str = Field(readOnly=True, description="Uniform resource identifier")
+    publishedBy: Union[HttpUrl, PublishActivity] = Field(
         description="The URL should contain the provenance of the publishing process.",
         readOnly=True,
         nskey="dandi",
     )  # TODO: formalize "publish" activity to at least the Actor
-    datePublished: date = Field(readOnly=True, nskey="schema")
+    datePublished: datetime = Field(readOnly=True, nskey="schema")
+    url: HttpUrl = Field(
+        readOnly=True, description="permalink to the item", nskey="schema"
+    )
+
+
+class PublishedDandisetMeta(DandisetMeta, Publishable):
+    version: str = Field(readOnly=True, nskey="schema")
+    doi: str = Field(
+        None,
+        title="DOI",
+        readOnly=True,
+        pattern=r"^10\.[A-Za-z0-9.\/-]+",
+        nskey="dandi",
+    )
+
+
+class PublishedAssetMeta(AssetMeta, Publishable):
+    pass
 
 
 def get_schema_version():
