@@ -1,4 +1,5 @@
 import pytest
+import responses
 
 from dandi.consts import known_instances
 from dandi.dandiarchive import follow_redirect, parse_dandi_url
@@ -129,6 +130,28 @@ def test_parse_girder_url(url, asset_type, asset_id):
         #         "folder_id": "5f176583f63d62e1dbd06943",
         #     },
         # )
+        (
+            "https://api.dandiarchive.org/api/dandisets/000003/versions/draft"
+            "/assets/0a748f90-d497-4a9c-822e-9c63811db412/download/",
+            "dandi-api",
+            "item",
+            {
+                "dandiset_id": "000003",
+                "version": "draft",
+                "asset_id": "0a748f90-d497-4a9c-822e-9c63811db412",
+            },
+        ),
+        (
+            "https://api.dandiarchive.org/api/dandisets/000003/versions/draft"
+            "/assets/?path=sub-YutaMouse20",
+            "dandi-api",
+            "folder",
+            {
+                "dandiset_id": "000003",
+                "version": "draft",
+                "location": "sub-YutaMouse20",
+            },
+        ),
     ],
 )
 def test_parse_api_url(url, instance, asset_type, asset_id):
@@ -152,3 +175,53 @@ def test_follow_redirect():
         follow_redirect("https://bit.ly/dandi12")
         == "https://gui.dandiarchive.org/#/file-browser/folder/5e72b6ac3da50caa9adb0498"
     )
+
+
+@responses.activate
+def test_parse_gui_old_redirect():
+    responses.add(
+        responses.GET,
+        "https://dandiarchive.org/server-info",
+        json={
+            "version": "1.2.0",
+            "cli-minimal-version": "0.6.0",
+            "cli-bad-versions": [],
+            "services": {
+                "girder": {"url": "https://girder.dandiarchive.org"},
+                "webui": {"url": "https://gui.dandirchive.org"},
+                "api": None,
+                "jupyterhub": {"url": "https://hub.dandiarchive.org"},
+            },
+        },
+    )
+    assert parse_dandi_url("https://gui.dandiarchive.org/#/dandiset/000003") == (
+        "girder",
+        "https://girder.dandiarchive.org/",
+        "dandiset",
+        {"dandiset_id": "000003", "version": "draft"},
+    )
+
+
+@responses.activate
+def test_parse_gui_new_redirect():
+    responses.add(
+        responses.GET,
+        "https://dandiarchive.org/server-info",
+        json={
+            "version": "1.2.0",
+            "cli-minimal-version": "0.6.0",
+            "cli-bad-versions": [],
+            "services": {
+                "girder": None,
+                "webui": {"url": "https://gui.dandirchive.org"},
+                "api": {"url": "https://api.dandiarchive.org/api"},
+                "jupyterhub": {"url": "https://hub.dandiarchive.org"},
+            },
+        },
+    )
+    assert parse_dandi_url("https://gui.dandiarchive.org/#/dandiset/000003") == (
+        "api",
+        "https://api.dandiarchive.org/api/",
+        "dandiset",
+        {"dandiset_id": "000003", "version": None},
+    )
